/*
 *  tab2sql - convert <TAB> delimited data to SQL
 *
 *  AUTHOR: Thomas Spahni (thomas@spahni.ch)
 *  Copyright (C) 2001 by the author
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef	_GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <getopt.h>

/*
 * Input is *.csv data which has been preprocessed with csv.c
 *
 * An attempt is made to create the proper CREATE clause.
 *
 */

#define TAB2SQL_VERSION	"1.0"
#define PROGNAME	"tab2sql"

#define MAXCOLUMNS	1024
#define FIELDBUFFER	40690
#define SOURCEBUFFER	1048576

#define TINYINT		1
#define SMALLINT	2
#define MEDIUMINT	4
#define INT		8
#define BIGINT		16
#define FLOAT		32
#define DOUBLE		64
#define CHAR		128
#define VARCHAR		256
#define TEXT		512


static char *columntype[] = {
	"illegal", "TINYINT", "SMALLINT", "MEDIUMINT", "INT", "BIGINT",
	"FLOAT", "DOUBLE", "CHAR(3)", "VARCHAR(255)", "TEXT"};

static struct option long_options[] = {
	{"exists",             no_argument,       0, 'e'},
	{"help",               no_argument,       0, 'h'},
	{"helpme",             no_argument,       0, '?'},
	{"table",              required_argument, 0, 't'},
	{"verbose",            no_argument,       0, 'v'},
	{"version",            no_argument,       0, 'V'},
	{0, 0, 0, 0}
};

int checkint(char *p) {
	char *q; char c;
	long i;

	q = p;

	/* the empty string is a tinyint */
	if(strlen(p) == 0) return(TINYINT);

	/* a sign as the first char is allowable for an int */
	if((*p == '+') || (*p == '-')) p++;

	while((c = *p++) != '\0') {
		if((c < '0') || (c > '9')) return(0);
	}

	/* seems to be an integer; get it */
	i = atol(q);
	if(i < 0) i *= -1;
	if(i <= 127) return(TINYINT);
	if(i <= 32767) return(SMALLINT);
	if(i <= 8388607) return(MEDIUMINT);
	if(i <= 2147483647) {
		return(INT);
	} else {
		return(0);
	}
}


int checkfloat(char *p) {
	char *q;
	char c;
	long double i;

	q = p;

	/* numbers 0-9, +/- and e/E is allowable for a float */
	while((c = *p++) != '\0') {
		if(! (((c > '0') && (c < '9')) ||
			(c == 'e') ||
			(c == 'E') ||
			(c == '+') ||
			(c == '-'))) return(0);
	}

	/* seems to be a float; get it */
	i = atof(q);
	if(i < 0) i *= -1;
	if(i <= 3.402823466E+38) return(FLOAT);
	if(i <= 1.7976931348623157E+308) {
		return(DOUBLE);
	} else {
		return(0);
	}
}


int checkstring(char *p) {
	size_t i;
	i = strlen(p);
	if(i <= 3) return(CHAR);
	if(i <= 255) return(VARCHAR);
	if(i <= 65535) {
		return(TEXT);
	} else {
		return(0);
	}
}


int checktype(char *p) {
	int datatype;
	if(datatype = checkint(p)) return(datatype);
	if(datatype = checkfloat(p)) return(datatype);
	if(datatype = checkstring(p)) return(datatype);
	fprintf(stderr, "[%s] ERROR: unrecognized datatype", PROGNAME);
	exit(1);
}


int striprest(int i) {
	int k = 0;
	while(i > 0) {
		i >>= 1;
		k++;
	}
	return(k);
}


static void print_version(void) {
	printf("%s Ver. %s\n\
Copyright (C) 2001 Thomas Spahni (GPL applies)\n",
	PROGNAME, TAB2SQL_VERSION);
}


static void usage(void) {
	printf("\n");
	print_version();
	printf("\n\
%s is a filter and takes its input from stdin.\n\
Input data consists of TAB delimited columns.\n\
Output is SQL code on stdout, ready to feed into a database.\n\n\
These options are supported:\n\
  -e  --exists     Create table with IF NOT EXISTS clause\n\
  -h  --help       Display this help and exit\n\
  -?  --helpme     Display this help and exit\n\
  -t  --table=...  the tablename given will be used\n\
                      instead of the default 'tab2sql'\n\
  -v  --verbose    Give some info on stderr while processing\n\
  -V  --version    Show version information and exit\n",
   PROGNAME);
}


/* --------------------------------------------------------*/
int main (int argc, char *argv[]) {

int   c, i, k, n, col;
int   escaped;
int   option_index = 0;
char *datapointer;
char *colnames;
int   ncols;
char *sbuf, *startsbuf;
char *startofbuffer;
char *columns[MAXCOLUMNS];
char *colpos[MAXCOLUMNS];
int   coltype[MAXCOLUMNS];

/* Variables for command line options */

int   exists  = 0;
char *table   = 0;
int   verbose = 0;

/* process command line options */
while ((c=getopt_long(argc, argv, "eh?vV:t",
		long_options, &option_index)) != EOF) {

	switch(c) {
	case 'e':
		exists = 1;
		break;
	case 'h':
	case '?':
		usage();
		exit(0);
		break;
	case 't':
		table = optarg;
		break;
	case 'v':
		verbose = 1;
		break;
	case 'V':
		print_version();
		exit(0);
		break;
	default:
		fprintf(stderr,
			"Illegal option character '%c'\n",
			opterr);
		break;
	}
}


if ((datapointer = malloc(FIELDBUFFER)) == NULL) {
	fprintf(stderr, "[%s] ERROR: could not allocate line buffer space\n",
		argv[0]);
	exit(1);
	}

if ((colnames = malloc(FIELDBUFFER)) == NULL) {
	fprintf(stderr, "[%s] ERROR: could not allocate column names space\n",
		argv[0]);
	exit(1);
	}

if ((startsbuf = malloc(SOURCEBUFFER)) == NULL) {
	fprintf(stderr, "[%s] ERROR: could not allocate input buffer space\n",
		argv[0]);
	exit(1);
	}


i = 0;
escaped      = 0;
colpos[i++]  = colnames;

/* here the first row od data is read from stdin
 * it contains a name for every column
 */
while ((c = getchar()) != '\n') {
	if ( (colnames - colpos[0]) >= FIELDBUFFER ) {
		fprintf(stderr, "[%s] ERROR: column name buffer overflow\n",
			PROGNAME);
		exit(1);
	}

	if ( i >= MAXCOLUMNS ) {
		fprintf(stderr, "[%s] ERROR: buffer overflow; more than %d columns\n", 
			PROGNAME, MAXCOLUMNS);
		exit(1);
	}

	if ((c == '\t') && (escaped == 0)) {
		*colnames++  = '\0';
		colpos[i++]  = colnames;
	} else { switch(c) {
		case ' ':
			*colnames++ = '_';
			break;
		case '.':
			*colnames++ = '_';
			break;
		case '/':
			*colnames++ = '_';
			break;
		default:
			*colnames++ = (char) c;
			break;
		}
	}

	if (c == '\\') {escaped = 1;} else {escaped = 0;}

} /* wend */
ncols = i;


/* --- now reading and checking datarows
 *     every field goes to type testing
 *     and the corresponding column's type is set.
 *
 *     limited testing for special characters or escaped characters
 *     is done.
 */

/* Initialize column types */
for ( k = 0; k < ncols; k++) {
	coltype[k] = 0;
}

startofbuffer = datapointer;
columns[i++]  = startofbuffer;
sbuf          = startsbuf;

/* escaped is a flag to indicate that the previous character
 * was the escape character;
 * i counts columns
 * n counts rows
 */
escaped  = 0;
i = 0;
n = 0;
columns[i++] = startofbuffer;


/* 
 * while getting all data from stdin a copy is put
 * into memory at 'startsbuf'.
 */

while ((c = getchar()) != EOF) {

	/* Source buffer may overflow; warn and exit */
	if ( (sbuf - startsbuf) >= SOURCEBUFFER ) {
		fprintf(stderr, "[%s] ERROR: source buffer overflow\n",
			PROGNAME);
		exit(1);
	}

	if ( (datapointer - columns[0]) > FIELDBUFFER ) {
		fprintf(stderr, "[%s] ERROR: line buffer overflow\n",
			PROGNAME);
		exit(1);
	}

	if ( i >= MAXCOLUMNS ) {
		fprintf(stderr, "[%s] ERROR: more than %d columns\n", 
			PROGNAME, MAXCOLUMNS);
		exit(1);
	}

	/* unescaped TAB is a field delimiter */
	if ((c == '\t') && (escaped == 0)) {
		*datapointer++ = '\0';
		columns[i++]   = datapointer;
		*sbuf++        = '\0';
	} else {
		if ((c != '\n') || (escaped == 1)) {
			*datapointer++ = (char) c;
			switch(c) {
			case '\n':
				*sbuf++ = '\\';
				*sbuf++ = 'n';
				break;
			case '\'':
				*sbuf++ = '\\';
				*sbuf++ = '\'';
				break;
			case '\\':
				*sbuf++ = '\\';
				*sbuf++ = '\\';
				break;
			case '\t':
				*sbuf++ = '\\';
				*sbuf++ = 't';
				break;
			default:
				*sbuf++        = (char) c;
				break;
			}
		}
	}

	if (c == '\\') {escaped = 1;} else {escaped = 0;}
	
	if ((c == '\n') && (escaped == 0)) {
		/* this is the end of a row */
		*datapointer++ = '\0';
		*sbuf++        = '\0';
		*sbuf++        = '\n';
		for ( k = 0; k < ncols; k++ ) {
			coltype[k] = coltype[k] |
				checktype(columns[k]);
		}

		if(i != ncols) {
			fprintf(stderr, "Data corrupt on line %d\n", n+1);
		}

		i = 0; n++;
		datapointer  = startofbuffer;
		columns[i++] = startofbuffer;
	}
} /* wend */

if(verbose) {
	printf("This was %d data rows to be read\n", n);
}

sbuf = startsbuf;
if ( table == 0 ) table = PROGNAME;

printf("# SQL code generated by %s from *.csv format\n\
#\n", PROGNAME);

printf("CREATE TABLE ");
if ( exists ) printf("IF NOT EXISTS ");
printf("%s (\n", table);

/* printf("We have %d columns\n", j); */
for ( k = 0; k < ncols; k++) {
	printf("  %s %s NOT NULL",
		colpos[k], columntype[striprest(coltype[k])]);
	if(k < ncols-1) {
		printf(",\n");
	} else {
		printf("\n);\n\n");
	}
}

printf("INSERT INTO %s VALUES\n", table);

sbuf = startsbuf;
i = 0;

for ( k = 0; k < n; k++) {
	printf("  (");
	columns[i++] = sbuf;
	/* get a row from the buffer */
	while((c = *sbuf++) != '\n') {
		if (c == '\0') {
			columns[i++] = sbuf;
		}
	}
	for(col = 0; col < ncols; col++) {
		if( striprest(coltype[col]) > 7) {
			printf("'%s'", columns[col]);
		} else {
			if(strlen(columns[col]) == 0) {
				printf("''");
			} else {
				printf("%s", columns[col]);
			}
		}

		if(col < (ncols-1)) {
			printf(",");
		}
	}
	i = 0;
	if(k < (n-1)) {
		printf("),\n");
	} else {
		printf(");\n\n");
	}
}


fflush(stdout);
exit(EXIT_SUCCESS);
}
